package org.processmining.pmind.algorithms;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.processmining.pmind.algorithms.concept.ModelComparison;
import org.processmining.pmind.algorithms.concept.PossibleModel;
import org.processmining.pmind.models.DMNActivityAttributePair;
import org.processmining.pmind.models.DMNDecisionActivity;
import org.processmining.pmind.models.DMNEdge;
import org.processmining.pmind.models.DMNModel;
import org.processmining.pmind.models.DMNNode;
import org.processmining.pmind.output.PMInDConformanceOutput;

import weka.classifiers.Evaluation;
import weka.core.Instances;

public class PMInDModelChecker {

	private AbstractDMNMiner output1,output2;
	private Set<DMNModel> unmatchedModels;
	private Set<ModelComparison> comparisons;
	private Set<ModelComparison> outcomes;
	
	public PMInDModelChecker(PMInDConformanceOutput input){
		output1 = input.getOuput1();
		output2 = input.getOuput2();
	}
	
	public void analyzeModels() throws Exception{		
		unmatchedModels = new HashSet<DMNModel>();
		comparisons = new HashSet<ModelComparison>();
		
		// Compare the second output vs. the "original" output 1
		for(DMNModel m: output2.getTopModels()){
			boolean f = false;
			for(DMNModel m2: output1.getTopModels()){				
				if(m.getTopActivity().weakEquals(m2.getTopActivity())){	
					ModelComparison outcome = checkModels(m, m2);
					
					outcome.addSubModels(doSubModels(m, m2.recursivelyGetSubModels()));
					//	outcome.addSubModels(outcomes);
					comparisons.add(outcome);
				}
			}
			if(!f)
				unmatchedModels.add(m);
		}
	}
	
	public Set<ModelComparison> doSubModels(DMNModel m, Set<DMNModel> m2) throws Exception{
		Set<ModelComparison> outcomes = new HashSet<ModelComparison>();
		
		for(DMNModel sm: m.getSubModels()){
			ModelComparison topModel = null;
			double bestScore = 0.0;
			for(DMNModel sm2: m2){
				ModelComparison outcome = checkModels(sm,sm2);
				if(outcome.getScore()>=bestScore){
					topModel = outcome;
					bestScore = outcome.getScore();
				}
			}
			if(bestScore>0){
				outcomes.add(topModel);
				topModel.addSubModels(doSubModels(sm, topModel.getModel2().recursivelyGetSubModels()));
			}
		}
		return outcomes;
	}
	
	public ModelComparison checkModels(DMNModel m, DMNModel m2) throws Exception{
		double contained = 0;
		
		Map<DMNActivityAttributePair,DMNActivityAttributePair> covered = new HashMap<DMNActivityAttributePair,DMNActivityAttributePair>();
		Set<DMNActivityAttributePair> missingInM = new HashSet<DMNActivityAttributePair>();
		Set<DMNActivityAttributePair> missingInM2 = new HashSet<DMNActivityAttributePair>();
		Set<DMNEdge> notConnectedInM1 = new HashSet<DMNEdge>();	
		Set<DMNEdge> notConnectedInM2 = new HashSet<DMNEdge>();	
		
		// Compare nodes
		missingInM2.addAll(m2.getAaPairs());
		for(DMNActivityAttributePair daap: m.getAaPairs()){
			boolean f = false;			
			for(DMNActivityAttributePair daap2: m2.getAaPairs()){
				if(daap.weakEquals(daap2)){
					covered.put(daap,daap2);					
					missingInM2.remove(daap2);
					f = true;
					break;
				}
			}
			if(!f)
				missingInM.add(daap);
		}		
		
		// Compare edges
		for(DMNEdge e: m.getEdges()){
			DMNDecisionActivity s = (DMNDecisionActivity) e.getSource();
			DMNDecisionActivity t = (DMNDecisionActivity) e.getTarget();
			DMNDecisionActivity s2 = null;
			DMNDecisionActivity t2 = null;
			for(DMNNode n: m2.getNodes()){
				if(n.getLabel().equals(s.getLabel()))
					s2 = (DMNDecisionActivity) n;
				if(n.getLabel().equals(t.getLabel()))
					t2 = (DMNDecisionActivity) n;
			}
			if(m2.areConnected(s2,t2)){
				double corrDiff = 1 - Math.abs(e.getCorr()- m2.getConnection(s2, t2).getCorr());
				contained+= corrDiff;
			}else 
				notConnectedInM1.add(e);
		}
		for(DMNEdge e: m2.getEdges()){
			DMNDecisionActivity s = (DMNDecisionActivity) e.getSource();
			DMNDecisionActivity t = (DMNDecisionActivity) e.getTarget();
			DMNDecisionActivity s2 = null;
			DMNDecisionActivity t2 = null;
			for(DMNNode n: m.getNodes()){
				if(n.getLabel().equals(s.getLabel()))
					s2 = (DMNDecisionActivity) n;
				if(n.getLabel().equals(t.getLabel()))
					t2 = (DMNDecisionActivity) n;
			}
			if(!m.areConnected(s2,t2))
				notConnectedInM2.add(e);			
				
		}
		double outcome = (contained+covered.size()) / (m.getEdges().size()+m.getAaPairs().size());
		outcome = Math.round(outcome * 100.0) / 100.0;
		ModelComparison comparison = new ModelComparison(
				m, m2, covered, missingInM, missingInM2, notConnectedInM1, notConnectedInM2, outcome);
		
		// Compare models
		for(DMNActivityAttributePair daap: covered.keySet()){
			boolean f = true;
			for(DMNActivityAttributePair daapIn: m.getInputsOfDAAP(daap)){
				if(covered.containsKey(daapIn)){
					if(!m2.areConnected(covered.get(daapIn), covered.get(daap),true)){
						f = false;
						break;
					}
				}else{
					f = false;
					break;
				}
			}		
			if(f && daap.getClassifier()!=null)
				comparison.addReplayResult(daap, Double.toString(checkReplayScore(m, daap, covered.get(daap))));
			else if (daap.getClassifier()!=null)
				comparison.addReplayResult(daap, "No replay possible due to missing inputs");
		}	
		return comparison;
	}
	
	private double checkReplayScore(DMNModel m, DMNActivityAttributePair d, DMNActivityAttributePair d2) throws Exception{
		double outcome = 0.0;
		if (d.weakEquals(d2)) {
			if (d2.getClassifier() != null) {
				//System.out.println("Doing :" + d + "'s traces over model of " + d2);

				PossibleModel pm = new PossibleModel(m.getInputsOfDAAP(d), d, d.getTraces());
				try {
					Instances in = output2.createTrainingSet(pm);
					Evaluation ev = new Evaluation(in);
					ev.crossValidateModel(d2.getClassifier(), in, 5, new Random(1));
					outcome = 1 - ev.rootMeanSquaredError();
				}
				catch (Exception e) {
					//System.out.println("This does not work :(");
				}
			}
		}	
		return outcome;
	}
	
	public Set<ModelComparison> getComparisons(){
		return comparisons;
	}
	
	public Set<DMNModel> getUnmatchedModels(){
		return unmatchedModels;
	}
	
}
