package org.processmining.pmind.plugins;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Toolkit;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.DotEdge;
import org.processmining.plugins.graphviz.dot.DotNode;
import org.processmining.plugins.graphviz.visualisation.DotPanel;
import org.processmining.plugins.inductiveVisualMiner.plugins.GraphvizPetriNet;
import org.processmining.pmind.algorithms.concept.TraceCluster;
import org.processmining.pmind.help.PMInDHelp;
import org.processmining.pmind.models.Activity;
import org.processmining.pmind.models.DMNActivityAttributePair;
import org.processmining.pmind.models.DMNDecisionActivity;
import org.processmining.pmind.models.DMNEdge;
import org.processmining.pmind.models.DMNModel;
import org.processmining.pmind.output.PMInDOutput;

public class PMInDOutputVisualizer {
	
	private boolean found = false;
	int scale = 5;
	UIPluginContext context;
	
	@Plugin(name = "P-MInD Decision Model Visualization",
			parameterLabels = { "P-MInD output"}, 
			returnLabels = { "DMN Models" }, 
			returnTypes = { JComponent.class }, 
			userAccessible = true,
			help = PMInDHelp.TEXT)
	
	@Visualizer
	@UITopiaVariant(affiliation = "University of Edinburgh",
			author = "Johannes De Smedt", 
			email = "johannes.desmedt@ed.ac.uk")
	public JComponent run(UIPluginContext context, PMInDOutput input) throws Exception {			
		this.context = context;
		boolean largeScreen = Toolkit.getDefaultToolkit().getScreenSize().getHeight()>1080;
		boolean perTrace = input.isPerTrace();
		
		if(largeScreen)
			scale = scale*2;
		
		final JPanel comp = new JPanel();
		comp.setLayout(new BorderLayout());

		JTabbedPane jtp = new JTabbedPane();
		if(largeScreen)	
			jtp.setFont(new Font("Helvetica", Font.PLAIN, 30));
		else
			jtp.setFont(new Font("Helvetica", Font.PLAIN, 15));
		
		// Visualize according to trace clusters
		
		// Every tracecluster gets its panel
		if(perTrace)
			for(TraceCluster tc: input.getTraceClusters()){
				JTabbedPane jtp2 = new JTabbedPane();
							
				for(DMNModel m: tc.getModels())
					addModelToPane(m, jtp2);
				
				jtp2.setPreferredSize(new Dimension(1000,1200));
				
				jtp.add(jtp2, ""+tc.getTraces().size());	
			}
		
		
		// Visualize according to model hierarchy
		if(!perTrace)
			for(DMNModel m: input.getModels()){
				if(m.getSuperModel()==null){
					JTabbedPane jtp2 = new JTabbedPane();
					
					addModelToPane(m, jtp2);
					addSubModels(m.getSubModels(),jtp2);
					
					jtp2.setPreferredSize(new Dimension(1000,1200));				
					jtp.add(jtp2, m.getLabel());	
				}
			}
		
				
		comp.add(jtp, BorderLayout.PAGE_START);
		return comp;
	}
	
	private void addSubModels(Set<DMNModel> models, JTabbedPane jtp2){		
		for(DMNModel m: models){
			addModelToPane(m, jtp2);
			addSubModels(m.getSubModels(),jtp2);
		}		
	}
	
	
	private void addModelToPane(DMNModel m, JTabbedPane jtp2){
		HashSet<Activity> activities = new HashSet<Activity>();
		HashSet<Activity> topActivities = new HashSet<Activity>();
		
		Collection<Marking> markings = new HashSet<Marking>();
		markings.add((Marking) m.getPetrinet()[2]);				
		
		JPanel p = new JPanel();
		p.setLayout(new GridLayout(2,1));
		
		// Add DMN model
		for(DMNActivityAttributePair daap: m.getAaPairs())
				activities.add(daap.getActivity());		
		topActivities.add(m.getTopActivity().getActivity());
		colorNodes(activities,topActivities,m.toDot());
		
		checkModelCompatibility(m, (Petrinet) m.getPetrinet()[0]);
		
		// Add DMN model
		DotPanel dotm = new DotPanel(m.toDot());
		dotm.setSize((int) dotm.getSize().getWidth()*scale*4, (int) dotm.getSize().getHeight()*scale*4);
		dotm.setPreferredSize(new Dimension((int) dotm.getSize().getWidth()*scale*2,(int) dotm.getSize().getHeight()*scale*2));
		
		p.add(dotm);
		
		// Add Petri net
		Dot d = GraphvizPetriNet.convert((Petrinet) m.getPetrinet()[0],(Marking) m.getPetrinet()[1], markings); 			
		colorNodes(activities,topActivities,d);
		
		DotPanel dp = new DotPanel(d);		
		
		p.add(dp);
		jtp2.add(p, m.getLabel());
	}
	
	
	
	private void checkModelCompatibility(DMNModel dmn, Petrinet pn){
		if(dmn.getTopActivity()!=null){
			DMNDecisionActivity top = dmn.getTopActivityNode();			
			checkPreviousNodes(dmn, pn, top);			
		}
	}
	
	private void checkPreviousNodes(DMNModel dmn, Petrinet pn, DMNDecisionActivity a){
		for (DMNEdge e : dmn.getInEdges(a)) {
			if(e.getSource() instanceof DMNDecisionActivity){
				DMNDecisionActivity da = (DMNDecisionActivity) e.getSource();
	
				// Color red if connection does not exist
				if (!connectionBetweenNodes(pn, a.getActivity(), da.getActivity()) && !a.getActivity().equals(da.getActivity())) {
					System.out.println("No connection between " + a + " and " + da);
					for (DotEdge dotEdge : dmn.toDot().getEdges()) {
						if ((dotEdge.getSource().getLabel().equals(da.toString())
								&& dotEdge.getTarget().getLabel().equals(a.toString()))
								|| (dotEdge.getSource().getLabel().equals(a.toString())
										&& dotEdge.getTarget().getLabel().equals(da.toString()))) {
							dotEdge.setOption("style", "filled");
							dotEdge.setOption("fillcolor", "#FF0000");
							dotEdge.setOption("color", "#FF0000");
							dotEdge.setOption("edgecolor", "#FF0000");
						}
					}
				}
				// Check previous nodes (remember, you start at top)
				checkPreviousNodes(dmn, pn, da);
			}
		}
	}
	
	private boolean connectionBetweenNodes(Petrinet pn, Activity a, Activity a2){
		Transition t1 = null, t2 = null;
		HashSet<PetrinetNode> visited = new HashSet<PetrinetNode>();
		
		for(PetrinetNode n: pn.getNodes()){
			if(n.getLabel().equals(a.getName()))
				t1 = (Transition) n;
			if(n.getLabel().equals(a2.getName()))
				t2 = (Transition) n;	
		}	
		if(t1==null || t2==null)
			return false;
		found = false;
		connectionBetweenTransitions(pn, t2, t1, visited);
		return found;
	}
	
	private void connectionBetweenTransitions(Petrinet pn, Transition t2, Transition inter,
			HashSet<PetrinetNode> visited) {
		if (!found && !visited.contains(inter)) {
			visited.add(inter);
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : pn.getInEdges(inter)) {
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e2 : pn.getInEdges(e.getSource())) {
					Transition t = (Transition) e2.getSource();
					if (t.equals(t2)) {
						found = true;
						break;
					}else 						
						connectionBetweenTransitions(pn, t2, t, visited);					
				}
			}
		}
	}
	
	
	public static void colorNodes(HashSet<Activity> activities, HashSet<Activity> topActivities, 
			Dot d){
		for(Activity a: activities){
			for(DotNode n: d.getNodes()){	
				if(a.getName().equals(n.getLabel())){
					n.setOption("style", "filled");
					n.setOption("fillcolor", "#8fffff");
				}
			}
		}
		for(Activity a: topActivities){
			for(DotNode n: d.getNodes()){	
				if(a.getName().equals(n.getLabel())){
					n.setOption("style", "filled");
					n.setOption("fillcolor", "#9D8FFF");
				}
			}
		}	
	}
			
}
